
(*** compound monads ***)

(** monad in the Kleisli category for M, 
  stuff exactly analogous to funM.ML, since compound monad is simply
  a monad in he Kleisli category for M **)

(* correspondence between this and an ordinary monad 
  o		oM
  id		unitM	
  unit		unitNM
  ext		pext
  oM		oNM
  map		kmap
  join		kjoin
  theorems : 
[ (id_o,	A1),
  (o_id,	A2),
  (o_assoc,	A3),
  (E2,		E2K),
  (E1,		E1K),
  (E3',		E3'K),
  (E3,		E3K),
  (E3'_oM,	E3'K_oNM),
  (E6,		E6K),
  (A5,		A5K),
  (E5,		E5K),
  (E4,		E4K),
  (ax8,		ax8K),
  (mapoext,	kmap_oM_pext),
  (is_ext',	is_pextK'),
  (is_ext,	is_pextK),
   ] ;
*) 

fun tr #"M" = "NM" | tr ch = BasisLibrary.String.str ch ;
fun M2NM str = BasisLibrary.String.translate tr str ;

signature STM = sig include THYSUB ;
  (* ordinary monad axioms and theorems *)
  structure M : MONAD ;
  end ;

signature UMC = sig 
  (* assumptions about unitNM and mapNM in compound monad *)
  val UC : thm	(* "unitNM == unitM o unitN" *)
  val MC : thm	(* "mapNM == mapM o mapN" *)
  end ;

(* monad in the Kleisli category of M, 3 axioms *)
signature MONADE123K = sig include STM ;
  val E2K : thm   (* "pext unitNM = unitM" *)
  val E1K : thm   (* "pext f oM unitNM = f" *)
  val E3K : thm(* "pext (g oNM f) = pext g oM pext f" *)
  end ;

signature E45K = sig 
  (* definitions of kmap and kjoin in terms of pext *)
  val E5K : thm ; (* "kmap f == pext (unitNM oM f)" *)
  val E4K : thm ; (* "kjoin == pext unitM" *)
  end ;

signature MONAD3K = sig include MONADE123K E45K ; end ;

(* monad in the Kleisli category of M, 7 axioms *)
signature SAX7 = sig 
  (* axioms for compound monad *)
  val sax1K : thm (* "kmap unitM = unitM" *)
  val sax2K : thm (* "kmap (g oM f) = kmap g oM kmap f" *)
  val sax3K : thm (* "kmap f oM unitNM = unitNM oM f" *)
  val sax4K : thm (* "kjoin oM kmap (kmap g) = kmap g oM kjoin" *)
  val sax5K : thm (* "kjoin oM unitNM = unitM" *)
  val sax6K : thm (* "kjoin oM kmap unitNM = unitM" *)
  val sax7K : thm (* "kjoin oM kmap kjoin = kjoin oM kjoin" *)
  end ;

signature MONAD7K = sig include STM SAX7 ; 
  (* definitions in terms of kjoin and kmap *)
  val ax8K : thm (* "pext f == kjoin oM kmap f" *)
  val E6K : thm (* "h oNM f == pext h oM f" *)
  end ;

(* compound monad with additional assumption A4K *)
signature MONADAK = sig include STM ;
  (* axioms re unitNM and oNM in Kleisli category of M *)
  val A2NM : thm 
  val A1NM : thm 
  val A3NM : thm 
  val A4K : thm 
  (* definition of pext in terms of oNM *)
  val A5K : thm 
  end ;

functor Kto3' (S3K : MONADE123K) : MONADE123 = struct local open S3K in
  val E1 = E1K ;
  val E2 = E2K ;
  val E3 = E3K ;
  val sub = sub ;
  val thy = thy ;
  structure Cs = M.Cs_oM ;
  end ; end ;

functor Kto3 (S3K : MONAD3K) : MONAD3 = struct 
  local structure S = Kto3' (S3K) in open S end ;
  val E4 = S3K.E4K ;
  val E5 = S3K.E5K ;
  end ;

functor Kto7 (S7K : MONAD7K) : MONAD7 = struct local open S7K in
  val ax1 = sax1K ;
  val ax2s = sax2K RS sym ;
  val ax3s = sax3K RS sym ;
  val ax4 = sax4K ;
  val ax5 = sax5K ;
  val ax6 = sax6K ;
  val ax7 = sax7K ;
  val ax8 = ax8K ;
  val E6 = E6K ;
  val sub = sub ;
  val thy = thy ;
  structure Cs = M.Cs_oM ;
  end ; end ;

functor KtoA (SAK : MONADAK) : MONADA = struct local open SAK in
  val A1 = A1NM ;
  val A2 = A2NM ;
  val A3 = A3NM ;
  val A4 = A4K ;
  val A5 = A5K ;
  val sub = sub ;
  val thy = thy ;
  structure Cs = M.Cs_oM ;
  end ; end ;

functor fromE123K (S3K : MONADE123K) = struct
local open S3K structure Stm : STM = S3K ; 
  structure MK3 = Kto3' (S3K) ;
in open Stm 
structure K : MONADo = fromE123 (MK3) ;

local open K in 
val E6K = E6 ;
val A5K = A5 ;
val A4K = A4 ;
val A3NM = A3 ;
val A1NM = A1 ;
val A2NM = A2 ;
end ;

val A6NM = (rep_trans [mks M.A4' RS K.Cs_oM.congR, K.A6, M.A4']) ;
val A4NM = (rep_trans [mks M.Cs.id_c RS K.Cs_oM.congR, A6NM]) ;

end ; (* local open S3K *)
end ; (* functor fromE123K *)

(* test that fromE123K gives MONADAK and MONADA axioms *)
(*
functor K3toAK (S3K : MONADE123K) : MONADAK = struct
  structure FE123K = fromE123K (S3K) ; open FE123K ; end ;
*)

(* show that the monad in the Kleisli category gives the 
  compound monad in the ordinary category, satisfying J1 *)
functor K3toA (structure S3K : MONADE123K ; 
  val A5NM : thm (* "extNM h == h oNM id" *)) =
struct local 
  structure FE123K = fromE123K (S3K) ; open FE123K ;
  structure NMA : MONADA = struct 
    val A3 = K.A3 ;
    val A2 = K.A2 ;
    val A1 = K.A1 ;
    val A4 = A4NM ;
    val A5 = A5NM ; 
    structure Cs = M.Cs ;
    val thy = thy ;
    val sub = M2NM ; end ;
  structure NM123 : MONADE123 = fromA (NMA) ;
in 
structure K = FE123K.K ;
structure M = S3K.M ;
structure NM : MONADo = fromE123 (NM123) ;
val J1Se = prove_goalw thy [M.E4, symmetric NM.E6, M.A5] 
  "extM (extNM f) = extNM f o joinM"
  (fn _ => [ (rewtac A5NM), (rtac (K.A6 RS sym) 1) ]) ;
  
val PE = rewrite_rule [NM.E6] K.A5 ;

val EC = prove_goalw thy [meta_eq_def] "extNM == extM o pext"
  (fn _ => [ (rtac ext 1), (simp_tac (etsimps thy [A5NM]) 1),
    (simp_tac (etsimps thy [M.A5, K.E6]) 1) ]) ;
    
end ; (* local *)
end ; (* functor K3toA *)

(* and conversely, given compound monad satisfying J1Se, using
  A5K as a definition, get monad in Kleisli category *)

functor J1toK (structure M : MONAD ; structure NM : MONADo ;
  val A5K : thm (* "pext h == h oNM unitM" *)
  val J1Se : thm (* "extM (extNM f) = extNM f o joinM" *)) : MONADAK = 
struct 
  structure M = M ;
  val A2NM = NM.A2 ;
  val A1NM = NM.A1 ;
  val A3NM = NM.A3 ;
  val A5K = A5K ;
  val thy = NM.thy ; fun sub s = s ;

val A4K = prove_sub_goalw sub thy [NM.E6, M.E6]
  "h oNM f = h oNM unitM oM f"
  (fn _ => [ (simp_tac (etsimps thy [M.extomap RS sym, J1Se]) 1),
    (simp_tac (etsimps thy [mk_oap M.ax6]) 1) ]) ;
end ; (* functor J1toK *)

(* structure S = fromE123K (S3K) ; *)

functor fromAK (SAK : MONADAK) : MONADE123K = struct
local open SAK structure Stm : STM = SAK ;
  structure MKA = KtoA (SAK) ;
  structure FMKA = fromAG (MKA) ;
in open Stm 

val E6K = FMKA.E6 ;
val E1K = FMKA.E1 ;
val E2K = FMKA.E2 ;
val E3K = FMKA.E3 ;
val A6K = FMKA.A6 ; 

end ; (* local open SAK *)
end ; (* functor fromAK *)

functor from3K (S3K : MONAD3K) = struct
local open S3K ; structure S = fromE123K (S3K) ; 
  structure MK3 = Kto3 (S3K) ;
in open S ;
  structure K = from3 (MK3) ;

local open K in 
val sax1K = ax1 ;
val sax2K = ax2 ;
val sax3K = ax3 ;
val sax5K = ax5 ;
val sax6K = ax6 ;
val sax4K = ax4 ;
val sax7K = ax7 ;
val sax4Ke = ax4e ;
val sax7Ke = ax7e ;
val ax8K = ax8 ;
end ;
end ; (* local open S3K *)
end ; (* functor from3K *)

(* test - gives system of 7 axioms from system of 3 axioms *)
functor K3to7 (S3K : MONAD3K) : MONAD7K = struct
  structure FS3K = from3K (S3K) ; open FS3K ; 
  val E6K = K.E6 ; end ;

(* 
structure Stm : STM = S3K ;
structure FS3K = from3K (S3K) ;
structure S7K : MONAD7K = struct open Stm FS3K end ;
structure FS7K = from7K (S7K) ;
structure SE123K : MONAD3K = struct open Stm FS7K end ;
structure FSE123K = from3K (SE123K) ;
*)

functor from7K (S7K : MONAD7K) : MONAD3K = struct
local open S7K structure S : STM = S7K ; 
  structure MK7 = Kto7 (S7K) ;
  structure FMK7 = from7 (MK7) ;
in open S ;

val E1K = FMK7.E1 ;
val E2K = FMK7.E2 ;
val E3K = FMK7.E3 ;
val E4K = FMK7.E4 ;
val E5K = FMK7.E5 ;

end ; (* local open S7K *)
end ; (* functor from7K *)

(*
structure MKA = KtoA (SAK) ;
structure FMKA = fromA (MKA) ;
structure MK7 = Kto7 (S7K) ;
structure FMK7 = from7G (MK7) ;
structure MK3 : MONAD3 = FMK7 ;
structure FMK3 = from3 (MK3) ;
*)

